/* Gauss-Jordan implementation */
/* Thanks to John Stearns for inspiration */

#include "2darray.h"
#include "stddefs.h"
#include <math.h>
#ifdef	UNIX
	#include	<ieeefp.h>
#endif
#ifdef	WIN32
	#include <float.h>
	#define	FP_PINF	_FPCLASS_PINF
#endif

/* -- Defines */

#define	SWAP(x, y)		{temp = (x); (x) = (y); (y) = temp;}

#define	GJ_SING_SENS	0.00					/* Singular matrix sensitivity */


/* -- Gauss-Jordan inverse */

bool GaussJordanInverse(double **matrix, double **inverse, int size)
{
	double	det,				/* Determinant (product of pivots) */
				temp,				/* Used by SWAP */
				pivot,			/* Current pivot value */
				factor;			/* Factor of pivot row to multiply by */
	int		ipass,			/* Elimination index (current pivot row) */
				imax,				/* Index of maximum value in pivot column */
				icol,				/* Column index */
				irow;				/* Row index */

	det = 1.0;

	/* Initially inverse is the identity matrix, */
	/* this will be replaced with the inverse    */
	for (irow = 0; irow < size; irow++) {
		for(icol = 0; icol < size; icol++) {
			if (irow == icol)
				inverse[irow][icol] = 1;
			else
				inverse[irow][icol] = 0;
		}
	}

	for (ipass = 0; ipass < size; ipass++) {
		/* Find maximum value in the pivot column */
		imax = ipass;
		for (irow = ipass; irow < size; irow++) {
			if (fabs(matrix[irow][ipass]) > fabs(matrix[imax][ipass]))
				imax = irow;
		}

		/* Exchange ipass row with imax row in both matricies */
		if(imax != ipass) {	
			for(icol = 0; icol < size; icol++) {
				SWAP(inverse[ipass][icol], inverse[imax][icol]);
				if(icol >= ipass)
					SWAP(matrix[ipass][icol], matrix[imax][icol]);
			}
		}

		/* Current pivot is now matrix[ipass][ipass] */
		/* Determinant is product of pivot elements */
		/* if det == 0 then matrix is singular and we can't invert */
		pivot = matrix[ipass][ipass];
		det *= pivot;
		if (fabs(det) <= GJ_SING_SENS) {
			/* Singular, so return infinite matrix */
			for (icol = 0; icol < size; icol++)
				for (irow = 0; irow < size; irow++)
					inverse[icol][irow] = FP_PINF;
			return FALSE;
		}

		/* Normalise pivot row by dividing by pivot element */
		for (icol = 0; icol < size; icol++) {
			inverse[ipass][icol] /= pivot;
			if (icol >= ipass)
				matrix[ipass][icol] /= pivot;
		}

		/* Add to each row a mutiple of the pivot row such that */
		/* the element of 'matrix' in the pivot column is zero  */
		for (irow = 0; irow < size; irow++) {
			if (irow != ipass) {
				factor = matrix[irow][ipass];
				for (icol = 0; icol < size; icol++) {
					inverse[irow][icol] -= factor * inverse[ipass][icol];
					matrix[irow][icol] -= factor * matrix[ipass][icol];
				}
			}
		}
	}
	return TRUE;
}


/* --- END of gauss2.c --- */
